from pytheus.metrics import Counter, Histogram, _Metric
from pytheus.registry import CollectorRegistry


class TestCollectorRegistry:
    def test_can_create_collector_registry(self):
        CollectorRegistry()

    def test_can_create_collector_registry_with_prefix(self):
        prefix = "myprefix"
        registry = CollectorRegistry(prefix=prefix)
        assert registry.prefix == prefix

    def test_register_metric(self):
        registry = CollectorRegistry()
        metric = _Metric("name", "desc", registry=None)
        registry.register(metric)

        assert metric.name in registry._collectors
        assert registry._collectors[metric.name] is metric._collector

    def test_register_metric_duplicate_name_is_ignored(self):
        registry = CollectorRegistry()
        metric_first = _Metric("name", "desc", registry=None)
        metric_second = _Metric("name", "desc", registry=None)
        registry.register(metric_first)
        registry.register(metric_second)

        assert metric_first is not metric_second
        assert registry._collectors[metric_first.name] is metric_first._collector

    def test_unregister_metric(self):
        registry = CollectorRegistry()
        metric = _Metric("name", "desc", registry=None)
        registry.register(metric)
        registry.unregister(metric)

        assert metric.name not in registry._collectors

    def test_collect(self):
        registry = CollectorRegistry()
        metric = Counter("name", "desc", registry=None)
        registry.register(metric)

        collectors = registry.collect()
        collectors = list(collectors)

        assert len(collectors) == 1

    def test_collect_with_multiple_metrics(self):
        registry = CollectorRegistry()
        counter = Counter("counter", "desc", registry=None)
        histogram = Histogram("histogram", "desc", registry=None)
        registry.register(counter)
        registry.register(histogram)

        collectors = registry.collect()
        collectors = list(collectors)

        assert len(collectors) == 2

    def test_collect_with_labeled_metrics(self):
        registry = CollectorRegistry()
        counter = Counter("name", "desc", required_labels=["bob", "cat"], registry=None)
        counter.labels({"bob": "2", "cat": "3"})
        registry.register(counter)

        collectors = registry.collect()
        collectors = list(collectors)

        assert len(collectors) == 1
        assert len(list(collectors[0].collect())) == 1

    def test_collect_with_labeled_metrics_unobservable(self):
        registry = CollectorRegistry()
        counter = Counter("name", "desc", required_labels=["bob", "cat"], registry=None)
        registry.register(counter)

        collectors = registry.collect()
        collectors = list(collectors)

        assert len(collectors) == 1
        assert len(list(collectors[0].collect())) == 0



from typing import TYPE_CHECKING

import redis

if TYPE_CHECKING:
    from pytheus.backends.base import BackendConfig
    from pytheus.metrics import _Metric


EXPIRE_KEY_TIME = 3600  # 1 hour


class MultiProcessRedisBackend:
    """
    Provides a multi-process backend that uses Redis.
    Single dimension metrics will be stored as list (ex. Counter) while labelled
    metrics will be stored in an hash.
    Currently is very naive, and there is a lot of room for improvement.
    (ex. maybe store all dimensions in the same hash and be smart on retrieving everything...)
    """

    CONNECTION_POOL: redis.Redis | None = None

    def __init__(
        self,
        config: "BackendConfig",
        metric: "_Metric",
        histogram_bucket: str | None = None,
    ) -> None:
        self._key_name = metric._collector.name
        self._labels_hash = None
        self._histogram_bucket = histogram_bucket

        # keys for histograms are of type `myhisto:2.5`
        if histogram_bucket:
            self._key_name = f"{self._key_name}:{histogram_bucket}"

        # default labels
        joint_labels = None
        if metric._collector._default_labels_count:
            joint_labels = metric._collector._default_labels.copy()  # type: ignore
            if metric._labels:
                joint_labels.update(metric._labels)

        if joint_labels:
            self._labels_hash = "-".join(sorted(joint_labels.values()))
        elif metric._labels:
            self._labels_hash = "-".join(sorted(metric._labels.values()))

        # initialize the key in redis
        self._init_key()

    @classmethod
    def _initialize(cls, config: "BackendConfig") -> None:
        cls.CONNECTION_POOL = redis.Redis(
            **config,
            decode_responses=True,
        )
        cls.CONNECTION_POOL.ping()

    def _init_key(self) -> None:
        """
        If the key doesn't exist in redis we initialize it and set the expiry time.
        `incrbyfloat` & `hincrbyfloat` are used so if for any reason multiple clients try to
        initialize the same key, it will be idempotent.
        """
        assert self.CONNECTION_POOL is not None
        if not self.CONNECTION_POOL.exists(self._key_name):
            if self._labels_hash:
                self.CONNECTION_POOL.hincrbyfloat(self._key_name, self._labels_hash, 0.0)
            else:
                self.CONNECTION_POOL.incrbyfloat(self._key_name, 0.0)

        self.CONNECTION_POOL.expire(self._key_name, EXPIRE_KEY_TIME)

    def inc(self, value: float) -> None:
        assert self.CONNECTION_POOL is not None
        if self._labels_hash:
            self.CONNECTION_POOL.hincrbyfloat(self._key_name, self._labels_hash, value)
        else:
            self.CONNECTION_POOL.incrbyfloat(self._key_name, value)

        self.CONNECTION_POOL.expire(self._key_name, EXPIRE_KEY_TIME)

    def dec(self, value: float) -> None:
        assert self.CONNECTION_POOL is not None
        if self._labels_hash:
            self.CONNECTION_POOL.hincrbyfloat(self._key_name, self._labels_hash, -value)
        else:
            self.CONNECTION_POOL.incrbyfloat(self._key_name, -value)

        self.CONNECTION_POOL.expire(self._key_name, EXPIRE_KEY_TIME)

    def set(self, value: float) -> None:
        assert self.CONNECTION_POOL is not None
        if self._labels_hash:
            self.CONNECTION_POOL.hset(self._key_name, self._labels_hash, value)
        else:
            self.CONNECTION_POOL.set(self._key_name, value)

        self.CONNECTION_POOL.expire(self._key_name, EXPIRE_KEY_TIME)

    def get(self) -> float:
        assert self.CONNECTION_POOL is not None
        if self._labels_hash:
            value = self.CONNECTION_POOL.hget(self._key_name, self._labels_hash)
        else:
            value = self.CONNECTION_POOL.get(self._key_name)

        if not value:
            self._init_key()
            return 0.0

        self.CONNECTION_POOL.expire(self._key_name, EXPIRE_KEY_TIME)
        return float(value)



class PytheusException(Exception):
    pass


class UnobservableMetricException(PytheusException):
    pass


class InvalidBackendClassException(PytheusException):
    pass


class InvalidBackendConfigException(PytheusException):
    pass


class LabelValidationException(PytheusException):
    pass


class BucketException(PytheusException):
    pass



from pytheus.utils import InfFloat


def test_inf_float_returns_correct_string():
    inf_float = InfFloat("inf")
    assert inf_float == float("inf")
    assert str(inf_float) == "+Inf"


def test_inf_float_with_wrong_value_still_returns_correct_string():
    inf_float = InfFloat(3)
    assert inf_float == 3.0
    assert str(inf_float) == "+Inf"



import importlib
import json
import os
from threading import Lock
from typing import TYPE_CHECKING, Any, Protocol, runtime_checkable

from pytheus.exceptions import InvalidBackendClassException, InvalidBackendConfigException

if TYPE_CHECKING:
    from pytheus.metrics import _Metric


BackendConfig = dict[str, Any]


@runtime_checkable
class Backend(Protocol):
    """
    Describes how to implement a Backend that can be used as a metric value backend.
    It must take the values indicated in the `__init__` method.
    """

    def __init__(
        # optional config ?
        self,
        config: BackendConfig,
        metric: "_Metric",
        histogram_bucket: str | None = None,
    ) -> None:
        ...

    def inc(self, value: float) -> None:
        ...

    def dec(self, value: float) -> None:
        ...

    def set(self, value: float) -> None:
        ...

    def get(self) -> float:
        ...


def _import_backend_class(full_import_path: str) -> type[Backend]:
    try:
        module_path, class_name = full_import_path.rsplit(".", 1)
    except ValueError:  # Empty string or not full path to the class
        raise InvalidBackendClassException(
            "Backend class could not be imported. Full import path needs to be provided, e.g. "
            "my_package.my_module.MyClass"
        )
    try:
        module = importlib.import_module(module_path)
    except ImportError as e:
        raise InvalidBackendClassException(f"Module '{module_path}' could not be imported: {e}")
    try:
        cls: type[Backend] = getattr(module, class_name)
        if not issubclass(cls, Backend):
            raise InvalidBackendClassException(f"Class '{class_name}' is not a Backend subclass")
        return cls
    except AttributeError:
        raise InvalidBackendClassException(
            f"Class '{class_name}' could not be found in module '{module_path}'"
        )


def load_backend(
    backend_class: type[Backend] | None = None,
    backend_config: BackendConfig | None = None,
) -> None:
    # Load default backend class
    global BACKEND_CLASS
    backend_class_env_var: str = "PYTHEUS_BACKEND_CLASS"
    if backend_class is not None:  # Explicit
        BACKEND_CLASS = backend_class
    elif backend_class_env_var in os.environ:  # Environment
        BACKEND_CLASS = _import_backend_class(os.environ[backend_class_env_var])
    else:  # Default
        BACKEND_CLASS = SingleProcessBackend

    # Load default backend config
    global BACKEND_CONFIG
    backend_config_env_var: str = "PYTHEUS_BACKEND_CONFIG"
    if backend_config is not None:  # Explicit
        BACKEND_CONFIG = backend_config
    elif backend_config_env_var in os.environ:  # Environment
        try:
            with open(os.environ[backend_config_env_var]) as f:
                BACKEND_CONFIG = json.loads(f.read())  # TODO: Support yaml?
        except Exception as e:
            raise InvalidBackendConfigException(f"{e.__class__}: {e}")
    else:
        BACKEND_CONFIG = {}  # Default

    # initialization hook
    if hasattr(BACKEND_CLASS, "_initialize"):
        BACKEND_CLASS._initialize(BACKEND_CONFIG)


def get_backend(metric: "_Metric", histogram_bucket: str | None = None) -> Backend:
    # Probably ok not to cache this and allow each metric to keep its own
    return BACKEND_CLASS(BACKEND_CONFIG, metric, histogram_bucket=histogram_bucket)


class SingleProcessBackend:
    """Provides a single-process backend that uses a thread-safe, in-memory approach."""

    def __init__(
        self,
        config: BackendConfig,
        metric: "_Metric",
        histogram_bucket: str | None = None,
    ) -> None:
        self._value = 0.0
        self._lock = Lock()

    def inc(self, value: float) -> None:
        with self._lock:
            self._value += value

    def dec(self, value: float) -> None:
        with self._lock:
            self._value -= value

    def set(self, value: float) -> None:
        with self._lock:
            self._value = value

    def get(self) -> float:
        with self._lock:
            return self._value


BACKEND_CLASS: type[Backend]
BACKEND_CONFIG: BackendConfig



import os
from typing import Callable

from pytheus.metrics import Labels
from pytheus.registry import REGISTRY, Collector, Registry

LINE_SEPARATOR = os.linesep
LABEL_SEPARATOR = ","
PROMETHEUS_CONTENT_TYPE = "text/plain; version=0.0.4; charset=utf-8"


def generate_metrics(registry: Registry = REGISTRY) -> str:
    """
    Returns the metrics from the registry in prometheus text format
    """
    lines = (
        generate_from_collector(collector, registry.prefix) for collector in registry.collect()
    )
    output = LINE_SEPARATOR.join(lines)
    output += "\n"
    return output


def format_labels(labels: Labels | None) -> str:
    if not labels:
        return ""
    label_str = (f'{name}="{value}"' for name, value in labels.items())
    return f"{{{LABEL_SEPARATOR.join(label_str)}}}"


def generate_from_collector(collector: Collector, prefix: str | None = None) -> str:
    """
    Returns the metrics from a given collector in prometheus text format
    """
    metric_name = f"{prefix}_{collector.name}" if prefix else collector.name
    help_text = f"# HELP {metric_name} {collector.description}"
    type_text = f"# TYPE {metric_name} {collector.type_}"
    output = [help_text, type_text]

    for sample in collector.collect():
        label_str = format_labels(sample.labels)
        metric = f"{metric_name}{sample.suffix}{label_str} {sample.value}"
        output.append(metric)
    return LINE_SEPARATOR.join(output)


def make_wsgi_app(registry: Registry = REGISTRY) -> Callable:
    """Create a WSGI app which serves the metrics from a registry."""

    def prometheus_app(environ, start_response):  # type: ignore
        status = "200 OK"
        if environ["PATH_INFO"] == "/favicon.ico":
            # Serve empty response for browsers
            headers = [("", "")]
            output = ""
        else:
            output = generate_metrics(registry)
            headers = [("Content-Type", PROMETHEUS_CONTENT_TYPE)]
        start_response(status, headers)
        return [output.encode()]

    return prometheus_app



from pytheus.exposition import generate_metrics
from pytheus.metrics import Counter, Histogram
from pytheus.registry import REGISTRY, CollectorRegistry


class TestExposition:
    def setup_counters(self):
        c = Counter("http_req_total", "metric desc", required_labels=["p", "m"])
        c.labels({"p": "p1", "m": "m1"}).inc(3)
        c.labels({"p": "p2", "m": "m2"}).inc(1)
        c = Counter("cache_hits_total", "cache desc", required_labels=["name"])
        c.labels({"name": "redis"}).inc(5)
        c = Counter("exception_total", "expection desc")
        c.inc(4)

    def test_generate_metrics(self):
        self.setup_counters()
        metrics_text = generate_metrics()
        assert metrics_text == (
            "# HELP http_req_total metric desc\n"
            "# TYPE http_req_total counter\n"
            'http_req_total{p="p1",m="m1"} 3.0\n'
            'http_req_total{p="p2",m="m2"} 1.0\n'
            "# HELP cache_hits_total cache desc\n"
            "# TYPE cache_hits_total counter\n"
            'cache_hits_total{name="redis"} 5.0\n'
            "# HELP exception_total expection desc\n"
            "# TYPE exception_total counter\n"
            "exception_total 4.0\n"
            ""
        )

    def test_generate_metrics_with_prefix(self):
        registry = CollectorRegistry(prefix="testing")
        REGISTRY.set_registry(registry)
        self.setup_counters()
        metrics_text = generate_metrics()
        assert metrics_text == (
            "# HELP testing_http_req_total metric desc\n"
            "# TYPE testing_http_req_total counter\n"
            'testing_http_req_total{p="p1",m="m1"} 3.0\n'
            'testing_http_req_total{p="p2",m="m2"} 1.0\n'
            "# HELP testing_cache_hits_total cache desc\n"
            "# TYPE testing_cache_hits_total counter\n"
            'testing_cache_hits_total{name="redis"} 5.0\n'
            "# HELP testing_exception_total expection desc\n"
            "# TYPE testing_exception_total counter\n"
            "testing_exception_total 4.0\n"
            ""
        )

    def test_generate_metrics_histogram(self):
        registry = CollectorRegistry()
        REGISTRY.set_registry(registry)

        histogram = Histogram("hello", "world")
        histogram.observe(0.4)
        metrics_text = generate_metrics()
        assert metrics_text == (
            '# HELP hello world\n# TYPE hello histogram\nhello_bucket{le="0.005"} 0.0\nhello_bucket{le="0.01"} 0.0\nhello_bucket{le="0.025"} 0.0\nhello_bucket{le="0.05"} 0.0\nhello_bucket{le="0.1"} 0.0\nhello_bucket{le="0.25"} 0.0\nhello_bucket{le="0.5"} 1.0\nhello_bucket{le="1"} 1.0\nhello_bucket{le="2.5"} 1.0\nhello_bucket{le="5"} 1.0\nhello_bucket{le="10"} 1.0\nhello_bucket{le="+Inf"} 1.0\nhello_sum 0.4\nhello_count 1.0\n'
        )

    def test_generate_metrics_histogram_with_labels(self):
        registry = CollectorRegistry()
        REGISTRY.set_registry(registry)

        histogram = Histogram("hello", "world", required_labels=["bob"])
        histogram_a = histogram.labels({"bob": "a"})
        histogram_b = histogram.labels({"bob": "b"})
        histogram_a.observe(0.4)
        histogram_b.observe(7)
        metrics_text = generate_metrics()
        assert metrics_text == (
            '# HELP hello world\n# TYPE hello histogram\nhello_bucket{bob="a",le="0.005"} 0.0\nhello_bucket{bob="a",le="0.01"} 0.0\nhello_bucket{bob="a",le="0.025"} 0.0\nhello_bucket{bob="a",le="0.05"} 0.0\nhello_bucket{bob="a",le="0.1"} 0.0\nhello_bucket{bob="a",le="0.25"} 0.0\nhello_bucket{bob="a",le="0.5"} 1.0\nhello_bucket{bob="a",le="1"} 1.0\nhello_bucket{bob="a",le="2.5"} 1.0\nhello_bucket{bob="a",le="5"} 1.0\nhello_bucket{bob="a",le="10"} 1.0\nhello_bucket{bob="a",le="+Inf"} 1.0\nhello_sum{bob="a"} 0.4\nhello_count{bob="a"} 1.0\nhello_bucket{bob="b",le="0.005"} 0.0\nhello_bucket{bob="b",le="0.01"} 0.0\nhello_bucket{bob="b",le="0.025"} 0.0\nhello_bucket{bob="b",le="0.05"} 0.0\nhello_bucket{bob="b",le="0.1"} 0.0\nhello_bucket{bob="b",le="0.25"} 0.0\nhello_bucket{bob="b",le="0.5"} 0.0\nhello_bucket{bob="b",le="1"} 0.0\nhello_bucket{bob="b",le="2.5"} 0.0\nhello_bucket{bob="b",le="5"} 0.0\nhello_bucket{bob="b",le="10"} 1.0\nhello_bucket{bob="b",le="+Inf"} 1.0\nhello_sum{bob="b"} 7.0\nhello_count{bob="b"} 1.0\n'
        )

    def test_generate_metrics_histogram_with_labels_and_default_labels(self):
        registry = CollectorRegistry()
        REGISTRY.set_registry(registry)

        histogram = Histogram(
            "hello", "world", required_labels=["bob"], default_labels={"bob": "default"}
        )
        histogram = histogram.labels({"bob": "a"})
        histogram.observe(0.4)
        metrics_text = generate_metrics()
        assert metrics_text == (
            '# HELP hello world\n# TYPE hello histogram\nhello_bucket{bob="a",le="0.005"} 0.0\nhello_bucket{bob="a",le="0.01"} 0.0\nhello_bucket{bob="a",le="0.025"} 0.0\nhello_bucket{bob="a",le="0.05"} 0.0\nhello_bucket{bob="a",le="0.1"} 0.0\nhello_bucket{bob="a",le="0.25"} 0.0\nhello_bucket{bob="a",le="0.5"} 1.0\nhello_bucket{bob="a",le="1"} 1.0\nhello_bucket{bob="a",le="2.5"} 1.0\nhello_bucket{bob="a",le="5"} 1.0\nhello_bucket{bob="a",le="10"} 1.0\nhello_bucket{bob="a",le="+Inf"} 1.0\nhello_sum{bob="a"} 0.4\nhello_count{bob="a"} 1.0\nhello_bucket{bob="default",le="0.005"} 0.0\nhello_bucket{bob="default",le="0.01"} 0.0\nhello_bucket{bob="default",le="0.025"} 0.0\nhello_bucket{bob="default",le="0.05"} 0.0\nhello_bucket{bob="default",le="0.1"} 0.0\nhello_bucket{bob="default",le="0.25"} 0.0\nhello_bucket{bob="default",le="0.5"} 0.0\nhello_bucket{bob="default",le="1"} 0.0\nhello_bucket{bob="default",le="2.5"} 0.0\nhello_bucket{bob="default",le="5"} 0.0\nhello_bucket{bob="default",le="10"} 0.0\nhello_bucket{bob="default",le="+Inf"} 0.0\nhello_sum{bob="default"} 0.0\nhello_count{bob="default"} 0.0\n'
        )



from enum import Enum


class MetricType(str, Enum):
    COUNTER = "counter"
    GAUGE = "gauge"
    HISTOGRAM = "histogram"
    SUMMARY = "summary"
    UNTYPED = "untyped"

    def __str__(self) -> str:
        return self.value


class InfFloat(float):
    """
    Overrides the float __str__ to return +Inf to align with the go/rust clients.
    This is expected to be used only as the upper bound in an Histogram bucket.
    """

    def __str__(self) -> str:
        """Avoiding checks as this is supposed to represent only float('inf')."""
        return "+Inf"



import time
from unittest import mock

import pytest

from pytheus.exceptions import (
    BucketException,
    LabelValidationException,
    UnobservableMetricException,
)
from pytheus.metrics import Counter, CustomCollector, Gauge, Histogram, _Metric, _MetricCollector
from pytheus.registry import REGISTRY, CollectorRegistry
from pytheus.utils import InfFloat, MetricType


@pytest.fixture
def set_empty_registry():
    """
    As the REGISTRY object is global by default, we might have data from other tests.
    So with this fixture we just set a new empty one.
    """
    REGISTRY.set_registry(CollectorRegistry())


class TestMetricCollector:
    @pytest.mark.parametrize(
        "name",
        [
            "prometheus_notifications_total",
            "process_cpu_seconds_total",
            "http_request_duration_seconds",
            "node_memory_usage_bytes",
            "http_requests_total",
            "foobar_build_info",
            "data_pipeline_last_record_processed_timestamp_seconds",
        ],
    )
    def test_name_with_correct_values(self, name):
        _MetricCollector(name, "desc", _Metric)

    @pytest.mark.parametrize(
        "name",
        [
            "invalid.name",
            "µspecialcharacter",
            "http_req@st_total",
            "http{request}",
        ],
    )
    def test_name_with_incorrect_values(self, name):
        with pytest.raises(ValueError):
            _MetricCollector(name, "desc", _Metric)

    def test_validate_required_labels_with_correct_values(self):
        labels = ["action", "method", "_type"]
        collector = _MetricCollector("name", "desc", _Metric)
        collector._validate_required_labels(labels)

    @pytest.mark.parametrize(
        "label",
        [
            "__private",
            "microµ",
            "@type",
        ],
    )
    def test_validate_required_labels_with_incorrect_values(self, label):
        collector = _MetricCollector("name", "desc", _Metric)
        with pytest.raises(LabelValidationException):
            collector._validate_required_labels([label])

    def test_collect_without_labels(self):
        counter = Counter("name", "desc")
        samples = counter._collector.collect()
        assert len(list(samples)) == 1

    def test_collect_with_labels(self):
        counter = Counter("name", "desc", required_labels=["a", "b"])
        counter_a = counter.labels({"a": "1", "b": "2"})  # noqa: F841
        counter_b = counter.labels({"a": "7", "b": "8"})  # noqa: F841
        counter_c = counter.labels({"a": "6"})  # this should not be creating a sample # noqa: F841
        samples = counter._collector.collect()
        assert len(list(samples)) == 2

    def test_collect_with_default_labels(self):
        counter = Counter("name", "desc", required_labels=["a"], default_labels={"a": 1})
        samples = counter._collector.collect()
        samples = list(samples)
        assert len(samples) == 1
        assert "a" in samples[0].labels
        assert 1 == samples[0].labels["a"]

    def test_collect_with_labels_and_default_labels(self):
        default_labels = {"a": 3}
        counter = Counter("name", "desc", required_labels=["a", "b"], default_labels=default_labels)
        counter_a = counter.labels({"a": "1", "b": "2"})  # noqa: F841
        counter_b = counter.labels({"a": "7", "b": "8"})  # noqa: F841
        counter_c = counter.labels({"a": "6"})  # this should not be creating a sample  # noqa: F841
        counter_d = counter.labels({"b": "5"})  # noqa: F841
        samples = counter._collector.collect()
        samples = list(samples)
        assert len(samples) == 3
        assert samples[0].labels["a"] == "1"
        assert samples[1].labels["a"] == "7"
        assert samples[2].labels["a"] == 3

    def test_collector_created_on_metric_creation(self):
        counter = Counter("name", "desc", required_labels=["a", "b"])
        assert counter._collector.name == "name"
        assert counter._collector.description == "desc"
        assert counter._collector._required_labels == {"a", "b"}

    def test_collector_reused_on_new_metric_instance(self):
        counter = Counter("name", "desc", required_labels=["a", "b"])
        counter_instance = Counter("name", "desc", collector=counter._collector)
        assert counter._collector is counter_instance._collector


class TestMetric:
    def test_create_metric(self):
        metric = _Metric("name", "desc")
        assert metric.name == "name"
        assert metric.description == "desc"
        assert metric.type_ == MetricType.UNTYPED

    def test_create_metric_without_registering_to_default_collector(self, set_empty_registry):
        metric = _Metric("name", "desc", registry=None)
        assert metric._registry is None
        assert metric._collector._registry is None
        assert len(list(REGISTRY.collect())) == 0

    def test_create_metric_with_required_labels(self):
        required_labels = ["bob", "cat"]
        metric = _Metric("name", "desc", required_labels=required_labels)
        assert metric._collector._required_labels == set(required_labels)

    def test_create_metric_raises_with_labels(self):
        # only default labels are allowed to be set on creation
        with pytest.raises(LabelValidationException):
            _Metric("name", "desc", required_labels=["a"], labels={"a": 1})

    def test_check_can_observe_without_required_labels(self):
        metric = _Metric("name", "desc")
        assert metric._check_can_observe() is True

    def test_check_can_observe_with_required_labels_without_labels(self):
        metric = _Metric("name", "desc", required_labels=["bob", "cat"])
        assert metric._check_can_observe() is False

    def test_check_can_observe_with_required_labels_with_partial_labels(self):
        metric = _Metric("name", "desc", required_labels=["bob", "cat"])
        metric = metric.labels({"bob": 2})
        assert metric._check_can_observe() is False

    def test_check_can_observe_with_required_labels_with_labels(self):
        metric = _Metric("name", "desc", required_labels=["bob", "cat"])
        metric = metric.labels({"bob": 1, "cat": 2})
        assert metric._check_can_observe() is True

    def test_raises_if_cannot_be_observed_observable(self):
        metric = _Metric("name", "desc", required_labels=["bob", "cat"])
        metric = metric.labels({"bob": 1, "cat": 2})
        metric._raise_if_cannot_observe()

    def test_raises_if_cannot_be_observed_unobservable(self):
        metric = _Metric("name", "desc", required_labels=["bob", "cat"])
        metric = metric.labels({"bob": 1})
        with pytest.raises(UnobservableMetricException):
            metric._raise_if_cannot_observe()

    def test_check_can_observe_with_default_labels(self):
        metric = _Metric(
            "name",
            "desc",
            required_labels=["bob", "cat"],
            default_labels={"bob": 1, "cat": 2},
        )
        assert metric._check_can_observe() is True

    def test_check_can_observe_with_default_labels_partial_uncomplete(self):
        metric = _Metric("name", "desc", required_labels=["bob", "cat"], default_labels={"bob": 1})
        assert metric._check_can_observe() is False

    def test_check_can_observe_with_default_labels_partial_complete(self):
        metric = _Metric("name", "desc", required_labels=["bob", "cat"], default_labels={"bob": 1})
        metric = metric.labels({"cat": 2})
        assert metric._check_can_observe() is True

    def test_check_can_observe_with_default_labels_partial_overriden_label(self):
        metric = _Metric("name", "desc", required_labels=["bob", "cat"], default_labels={"bob": 1})
        metric = metric.labels({"cat": 2, "bob": 2})
        assert metric._check_can_observe() is True

    # labels

    def test_labels_without_labels_return_itself(self):
        metric = _Metric("name", "desc")
        new = metric.labels({})
        assert new is metric

    def test_labels_without_required_labels_raises(self):
        metric = _Metric("name", "desc")
        with pytest.raises(LabelValidationException):
            metric.labels({"a": 1})

    def test_labels_unobservable(self):
        metric = _Metric("name", "desc", required_labels=["a", "b"])
        metric = metric.labels({"a": 1})
        assert metric not in metric._collector._labeled_metrics.values()

    def test_labels_observable(self):
        metric = _Metric("name", "desc", required_labels=["a", "b"])
        metric = metric.labels({"a": 1, "b": 2})
        assert metric in metric._collector._labeled_metrics.values()

    def test_labels_observable_returns_existing_child(self):
        metric = _Metric("name", "desc", required_labels=["a", "b"])
        metric_a = metric.labels({"a": 1, "b": 2})
        metric_b = metric.labels({"a": 1, "b": 2})
        assert len(metric._collector._labeled_metrics) == 1
        assert metric_a is metric_b

    def test_labels_with_unknown_label(self):
        metric = _Metric("name", "desc", required_labels=["a", "b"])
        with pytest.raises(LabelValidationException):
            metric.labels({"a": 1, "c": 2})

    # default_labels

    def test_metric_with_default_labels(self):
        default_labels = {"bob": "bobvalue"}
        metric = _Metric("name", "desc", required_labels=["bob"], default_labels=default_labels)
        assert metric._collector._default_labels == default_labels

    def test_metric_with_default_labels_raises_without_required_labels(self):
        default_labels = {"bob": "bobvalue"}
        with pytest.raises(LabelValidationException):
            _Metric("name", "desc", default_labels=default_labels)

    def test_metric_with_default_labels_with_label_not_in_required_labels(self):
        default_labels = {"bobby": "bobbyvalue"}
        with pytest.raises(LabelValidationException):
            _Metric("name", "desc", required_labels=["bob"], default_labels=default_labels)

    def test_metric_with_default_labels_with_subset_of_required_labels(self):
        default_labels = {"bob": "bobvalue"}
        metric = _Metric(
            "name",
            "desc",
            required_labels=["bob", "bobby"],
            default_labels=default_labels,
        )
        assert metric._collector._default_labels == default_labels

    def test_get_sample(self):
        from pytheus.metrics import Sample

        metric = _Metric("name", "desc")
        sample = metric._get_sample()
        assert sample == Sample("", None, 0)

    def test_add_default_labels_to_sample(self):
        default_labels = {"bob": "bobvalue"}
        metric = _Metric(
            "name",
            "desc",
            required_labels=["bob", "cat"],
            default_labels=default_labels,
        )
        metric = metric.labels({"cat": 2})
        sample = metric._get_sample()

        assert sample.labels == {"bob": "bobvalue", "cat": 2}

    def test_add_default_labels_to_sample_does_not_ovveride_provided_labels(self):
        default_labels = {"bob": "bobvalue"}
        metric = _Metric(
            "name",
            "desc",
            required_labels=["bob", "cat"],
            default_labels=default_labels,
        )
        metric = metric.labels({"cat": 2, "bob": "newvalue"})
        sample = metric._get_sample()

        assert sample.labels == {"bob": "newvalue", "cat": 2}


class TestCounter:
    @pytest.fixture
    def counter(self):
        return Counter("name", "desc")

    def test_metric_type(self, counter):
        assert counter.type_ == MetricType.COUNTER

    def test_can_increment(self, counter):
        counter.inc()
        assert counter._metric_value_backend.get() == 1

    def test_can_increment_with_value(self, counter):
        counter.inc(7.2)
        assert counter._metric_value_backend.get() == 7.2

    def test_negative_increment_raises(self, counter):
        with pytest.raises(ValueError):
            counter.inc(-1)

    def test_count_exception(self, counter):
        with pytest.raises(ValueError):
            with counter.count_exceptions():
                raise ValueError

        assert counter._metric_value_backend.get() == 1

    def test_count_exception_with_specified(self, counter):
        with pytest.raises(ValueError):
            with counter.count_exceptions((IndexError, ValueError)):
                raise ValueError

        assert counter._metric_value_backend.get() == 1

    def test_count_exception_with_specified_is_ignored(self, counter):
        with pytest.raises(ValueError):
            with counter.count_exceptions(IndexError):
                raise ValueError

        assert counter._metric_value_backend.get() == 0

    def test_count_exception_with_decorator(self, counter):
        @counter
        def test():
            raise ValueError

        with pytest.raises(ValueError):
            test()

        assert counter._metric_value_backend.get() == 1

    def test_count_exception_with_decorator_multiple(self, counter):
        @counter
        def test():
            raise ValueError

        @counter
        def test_2():
            raise ValueError

        with pytest.raises(ValueError):
            test()
        with pytest.raises(ValueError):
            test_2()

        assert counter._metric_value_backend.get() == 2

    def test_count_exception_with_specified_as_decorator(self, counter):
        @counter(exceptions=(IndexError, ValueError))
        def test():
            raise ValueError

        with pytest.raises(ValueError):
            test()

        assert counter._metric_value_backend.get() == 1

    def test_count_exception_with_specified_is_ignored_as_decorator(self, counter):
        @counter(exceptions=IndexError)
        def test():
            raise ValueError

        with pytest.raises(ValueError):
            test()

        assert counter._metric_value_backend.get() == 0


class TestGauge:
    @pytest.fixture
    def gauge(self):
        return Gauge("name", "desc")

    def test_metric_type(self, gauge):
        assert gauge.type_ == MetricType.GAUGE

    def test_gauge_starts_at_zero(self, gauge):
        assert gauge._metric_value_backend.get() == 0

    def test_can_increment(self, gauge):
        gauge.inc()
        assert gauge._metric_value_backend.get() == 1

    def test_can_increment_with_value(self, gauge):
        gauge.inc(7.2)
        assert gauge._metric_value_backend.get() == 7.2

    def test_can_increment_with_negative_value(self, gauge):
        gauge.inc(-7.2)
        assert gauge._metric_value_backend.get() == -7.2

    def test_can_decrement(self, gauge):
        gauge.dec()
        assert gauge._metric_value_backend.get() == -1

    def test_can_decrement_with_value(self, gauge):
        gauge.dec(7.2)
        assert gauge._metric_value_backend.get() == -7.2

    def test_can_decrement_with_negative_value(self, gauge):
        gauge.dec(-7.2)
        assert gauge._metric_value_backend.get() == 7.2

    def test_can_set(self, gauge):
        gauge.set(5.2)
        assert gauge._metric_value_backend.get() == 5.2

    def test_can_set_negative(self, gauge):
        gauge.set(-5.2)
        assert gauge._metric_value_backend.get() == -5.2

    # this has to fail sooner or later, the question is when ?
    def test_set_to_current_time(self, gauge):
        time_ = time.time()
        gauge.set_to_current_time()
        assert int(gauge._metric_value_backend.get()) == int(time_)

    def test_track_inprogress(self, gauge):
        with gauge.track_inprogress():
            assert gauge._metric_value_backend.get() == 1

    def test_track_inprogress_multiple(self, gauge):
        # will increment and decrement when exiting the context manager
        with gauge.track_inprogress():
            pass

        with gauge.track_inprogress():
            with gauge.track_inprogress():
                assert gauge._metric_value_backend.get() == 2

    def test_time(self, gauge):
        with gauge.time():
            pass
        assert gauge._metric_value_backend.get() != 0

    def test_as_decorator(self, gauge):
        @gauge
        def test():
            pass

        test()
        assert gauge._metric_value_backend.get() != 0

    def test_as_decorator_with_track_inprogress(self, gauge):
        backend_mock = mock.Mock()
        gauge._metric_value_backend = backend_mock

        @gauge(track_inprogress=True)
        def test():
            pass

        test()
        backend_mock.inc.assert_called()
        backend_mock.dec.assert_called()

    def test_as_decorator_with_track_inprogress_as_false(self, gauge):
        @gauge(track_inprogress=False)
        def test():
            pass

        test()
        assert gauge._metric_value_backend.get() != 0


class TestHistogram:
    @pytest.fixture
    def histogram(self):
        return Histogram("name", "desc")

    def test_metric_type(self, histogram):
        assert histogram.type_ == MetricType.HISTOGRAM

    def test_create_histogram_with_default_labels(self):
        Histogram("name", "desc", required_labels=["bob", "cat"])

    def test_histogram_fails_with_le_label(self):
        with pytest.raises(LabelValidationException):
            Histogram("name", "desc", required_labels=["bob", "le"])

    def test_buckets_adds_inf_implicitly_float_number(self):
        buckets = [0.2, 0.5, 1]
        expected = [0.2, 0.5, 1, InfFloat("inf")]
        histogram = Histogram("name", "desc", buckets=buckets)
        assert histogram._upper_bounds == expected
        assert isinstance(histogram._upper_bounds[-1], InfFloat)

    def test_buckets_adds_inf_implicitly_float_inf(self):
        buckets = [0.2, 0.5, 1, float("inf")]
        expected = [0.2, 0.5, 1, InfFloat("inf")]
        histogram = Histogram("name", "desc", buckets=buckets)
        assert histogram._upper_bounds == expected
        assert isinstance(histogram._upper_bounds[-1], InfFloat)

    def test_buckets_adds_inf_implicitly_inf_float(self):
        buckets = [0.2, 0.5, 1, InfFloat("inf")]
        expected = [0.2, 0.5, 1, InfFloat("inf")]
        histogram = Histogram("name", "desc", buckets=buckets)
        assert histogram._upper_bounds == expected
        assert isinstance(histogram._upper_bounds[-1], InfFloat)

    def test_buckets_with_sorted_order(self):
        buckets = [0.2, 0.5, 1]
        histogram = Histogram("name", "desc", buckets=buckets)
        assert histogram._upper_bounds == buckets + [InfFloat("inf")]

    def test_buckets_with_unsorted_order_fails(self):
        with pytest.raises(BucketException):
            Histogram("name", "desc", buckets=(0.2, 1, 0.5))

    def test_buckets_empty_uses_default_buckets(self):
        histogram = Histogram("name", "desc", buckets=[])
        assert histogram._upper_bounds == list(histogram.DEFAULT_BUCKETS) + [float("inf")]

    def test_does_not_have_metric_value_backend(self, histogram):
        assert histogram._metric_value_backend is None

    def test_unobservable_does_not_create_buckets(self):
        histogram = Histogram("name", "desc", required_labels=["bob"])
        assert histogram._buckets is None
        assert histogram._sum is None
        assert histogram._count is None

    def test_observable_creates_buckets(self):
        histogram = Histogram("name", "desc", required_labels=["bob"])
        histogram = histogram.labels({"bob": "cat"})
        assert histogram._sum is not None
        assert histogram._count is not None
        assert histogram._buckets is not None
        assert len(histogram._buckets) == len(histogram._upper_bounds)

    def test_labeled_observable_respects_custom_buckets(self):
        buckets = [0.2, 0.5, 1]
        histogram = Histogram("name", "desc", required_labels=["bob"], buckets=buckets)
        child_histogram = histogram.labels({"bob": "cat"})
        assert child_histogram._upper_bounds == histogram._upper_bounds
        assert len(child_histogram._buckets) == len(histogram._upper_bounds)

    def test_collect(self):
        buckets = [0.2, 0.5, 1]
        histogram = Histogram("name", "desc", buckets=buckets)
        samples = histogram.collect()
        samples = list(samples)

        assert "le" in samples[0].labels
        assert "0.2" in samples[0].labels.values()
        assert len(samples) == 6  # includes float('inf')

    def test_osberve_unobservable_raises(self):
        histogram = Histogram("name", "desc", required_labels=["bob"])
        with pytest.raises(UnobservableMetricException):
            histogram.observe(2)

    def test_observe(self):
        histogram = Histogram("name", "desc", buckets=[0.2, 0.5, 1])
        histogram.observe(0.4)

        assert histogram._sum.get() == 0.4
        assert histogram._count.get() == 1
        assert histogram._buckets[0].get() == 0
        assert histogram._buckets[1].get() == 1
        assert histogram._buckets[2].get() == 1

    def test_time(self, histogram):
        with histogram.time():
            pass
        assert histogram._count.get() == 1
        assert histogram._count.get() != 0

    def test_as_decorator(self, histogram):
        @histogram
        def test():
            pass

        test()
        assert histogram._count.get() == 1
        assert histogram._count.get() != 0

    def test_as_decorator_multiple(self, histogram):
        @histogram
        def test():
            pass

        @histogram
        def test_2():
            pass

        test()
        test_2()
        assert histogram._count.get() == 2
        assert histogram._count.get() != 0


class _TestCollector(CustomCollector):
    def collect(self):
        counter = Counter("name", "desc", registry=None)
        counter.inc()
        yield counter


class TestCustomCollector:
    def test_create_custom_collector(self):
        registry = CollectorRegistry()
        registry.register(_TestCollector())

        assert "_testcollector" in registry._collectors

    def test_custom_collector_collect(self):
        custom_collector = _TestCollector()
        metrics = list(custom_collector.collect())

        assert len(metrics) == 1
        assert metrics[0]._metric_value_backend._value == 1.0

    def test_cannot_add_two_custom_collectors_with_same_name(self, set_empty_registry):
        first_collector = _TestCollector()
        second_collector = _TestCollector()
        registry = CollectorRegistry()
        registry.register(first_collector)
        registry.register(second_collector)

        assert first_collector is not second_collector
        assert registry._collectors["_testcollector"] is first_collector

    def test_registry_correctly_return_custom_collector(self):
        custom_collector = _TestCollector()
        registry = CollectorRegistry()
        registry.register(custom_collector)

        collectors = registry.collect()
        assert list(collectors)[0] is custom_collector

    def test_can_remove_from_registry(self):
        custom_collector = _TestCollector()
        registry = CollectorRegistry()
        registry.register(custom_collector)
        registry.unregister(custom_collector)

        collectors = registry.collect()
        assert len(list(collectors)) == 0

    def test_adding_to_registry_with_already_present_metric_with_same_name(self):
        registry = CollectorRegistry()
        counter = Counter("name", "desc", registry=registry)
        registry.register(_TestCollector())

        assert len(registry._collectors) == 1
        assert registry._collectors["name"] is counter._collector



import functools
import itertools
import re
import time
from collections.abc import Generator
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Callable, Iterable, Sequence

from pytheus.backends import get_backend
from pytheus.exceptions import (
    BucketException,
    LabelValidationException,
    UnobservableMetricException,
)
from pytheus.registry import REGISTRY, Collector, Registry
from pytheus.utils import InfFloat, MetricType

Labels = dict[str, str]


metric_name_re = re.compile(r"[a-zA-Z_:][a-zA-Z0-9_:]*")
label_name_re = re.compile(r"[a-zA-Z_][a-zA-Z0-9_]*")


@dataclass
class Sample:
    suffix: str
    labels: dict[str, str] | None
    value: float


class CustomCollector(Collector):
    """
    Inheriting from this protocol is the current way to create a custom collector.
    The property `name` will return the lowercased class name so that it can be
    tracked in a registry.
    Not particularly liking this solution but for now it should suffice due to
    the current structure.
    """

    @property
    def name(self) -> str:
        return str.lower(self.__class__.__name__)

    # overriding the Collector protocol it requires to be writeable so we have this dummy setter
    @name.setter
    def name(self, value: str) -> None:
        pass


class _MetricCollector:
    """
    This handles the core logic for a specific named metric.
    It will do label validation, it will have the metric name and most importantly it will keep
    track of all the Child instances so that when `collect()` is called it will retrieve all
    the correct samples from all the possible observable Childrens.
    _labeled_metrics contains all observable metrics when required_labels is set.
    """

    def __init__(
        self,
        name: str,
        description: str,
        metric: "_Metric",
        required_labels: Sequence[str] | None = None,
        default_labels: Labels | None = None,
        registry: Registry | None = REGISTRY,
    ) -> None:
        if metric_name_re.fullmatch(name) is None:
            raise ValueError(f"Invalid metric name: {name}")

        if required_labels:
            is_histogram = isinstance(metric, Histogram)
            self._validate_required_labels(required_labels, is_histogram)

        self._required_labels = set(required_labels) if required_labels else None

        if default_labels:
            self._validate_labels(default_labels)

        self.name = name
        self.description = description
        self.type_ = metric.type_
        self._default_labels = default_labels
        self._default_labels_count = len(default_labels) if default_labels else 0
        self._metric = metric
        self._labeled_metrics: dict[tuple[str, ...], _Metric] = {}
        self._registry = registry

        if registry:
            registry.register(self)

    def _validate_required_labels(self, labels: Sequence[str], is_histogram: bool = False) -> None:
        """
        Validates label names according to the regex.
        Labels starting with `__` are reserved for internal use by Prometheus.
        """
        for label in labels:
            if label.startswith("__") or label_name_re.fullmatch(label) is None:
                raise LabelValidationException(f"Invalid label name: {label}")
            if is_histogram and label == "le":
                raise LabelValidationException(f"Invalid label name for Histogram: {label}")

    def _validate_labels(self, labels: Labels) -> None:
        """
        Validates labels.
        If no `required_labels` is set raises an error.
        `labels` will be also validated to be a subset of `required_labels`.
        """
        if not self._required_labels:
            raise LabelValidationException("trying to use labels while required_labels is None")

        labels_set = set(labels.keys())
        if not labels_set.issubset(self._required_labels):
            raise LabelValidationException(
                "labels different than required_labels: " f"{labels_set} != {self._required_labels}"
            )

    def collect(self) -> Iterable[Sample]:
        """
        Collects Samples for the metric.
        """
        labeled_metrics: Iterable[Sample]
        if self._required_labels:
            labeled_metrics = (
                sample for metric in self._labeled_metrics.values() for sample in metric.collect()
            )
            if self._default_labels and self._metric._can_observe:
                labeled_metrics = itertools.chain(labeled_metrics, self._metric.collect())
            return labeled_metrics
        else:
            return self._metric.collect()


class _Metric:
    type_: MetricType = MetricType.UNTYPED

    def __init__(
        self,
        name: str,
        description: str,
        required_labels: Sequence[str] | None = None,
        labels: Labels | None = None,
        default_labels: Labels | None = None,
        registry: Registry | None = REGISTRY,
        collector: _MetricCollector | None = None,
    ) -> None:
        self.name = name
        self.description = description
        self._labels = labels
        self._registry = registry
        self._metric_value_backend = None
        self._collector = (
            collector
            if collector
            else _MetricCollector(
                name, description, self, required_labels, default_labels, registry
            )
        )
        self._can_observe = self._check_can_observe()

        if not collector and labels:
            raise LabelValidationException(
                "Setting labels when creating a metric is not allowed. "
                "You might be looking for default_labels."
            )

        if self._can_observe and not isinstance(self, Histogram):
            self._metric_value_backend = get_backend(self)

    def _check_can_observe(self) -> bool:
        if not self._collector._required_labels:
            return True

        required_labels_count = len(self._collector._required_labels)
        if (
            self._collector._default_labels_count
            and self._collector._default_labels_count == required_labels_count
        ):
            return True

        if not self._labels:
            return False

        if self._collector._default_labels:
            default_labels = self._collector._default_labels.copy()
            default_labels.update(self._labels)
            labels_count = len(default_labels)
        else:
            labels_count = len(self._labels)

        if labels_count != required_labels_count:
            return False

        return True

    def _raise_if_cannot_observe(self) -> None:
        """Raise if the metric cannot be observed, for example if labels values are missing."""
        if not self._can_observe:
            raise UnobservableMetricException

    def labels(self, labels_: Labels) -> "_Metric":
        """
        If no labels is passed to the call returns itself.
        If there are already present labels, they will be updated with the passed labels_ and if
        it's not observable will just return the new child instance. Otherwise it will also add
        the instance to the labeled_metrics on the collector.
        """
        # if not labels_ or self._collector._required_labels is None:
        if not labels_:
            return self

        self._collector._validate_labels(labels_)

        if self._labels:
            new_labels = self._labels.copy()
            new_labels.update(labels_)
            labels_ = new_labels

        if self._collector._default_labels:
            default_labels = self._collector._default_labels.copy()
            default_labels.update(labels_)
            labels_count = len(default_labels)
        else:
            labels_count = len(labels_)

        # __init__ arguments for the child
        child_kwargs = {
            "name": self.name,
            "description": self.description,
            "collector": self._collector,
            "labels": labels_,
            "registry": self._registry,
        }

        # pass down buckets list to the child
        if isinstance(self, Histogram):
            child_kwargs["buckets"] = self._upper_bounds

        assert self._collector._required_labels is not None
        if labels_count != len(self._collector._required_labels):
            # does not add to collector
            return self.__class__(**child_kwargs)  # type: ignore

        # add to collector
        sorted_label_values = tuple(v for _, v in sorted(labels_.items()))
        if sorted_label_values in self._collector._labeled_metrics:
            metric = self._collector._labeled_metrics[sorted_label_values]
        else:
            metric = self.__class__(**child_kwargs)  # type: ignore
            self._collector._labeled_metrics[sorted_label_values] = metric
        return metric

    def collect(self) -> Iterable[Sample]:
        raise NotImplementedError

    def _get_sample(self) -> Sample:
        """Get a Sample for testing. Each metric type will need to define its own."""
        self._raise_if_cannot_observe()
        sample = Sample("", self._labels, 0)
        return self._add_default_labels_to_sample(sample)

    def _add_default_labels_to_sample(self, sample: Sample) -> Sample:
        """Adds default labels if available to a sample."""
        if self._collector._default_labels_count:
            joint_labels = self._collector._default_labels.copy()  # type: ignore
            if sample.labels:
                joint_labels.update(sample.labels)
            sample.labels = joint_labels

        return sample

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}({self._collector.name})"


class Counter(_Metric):
    type_: MetricType = MetricType.COUNTER

    def inc(self, value: float = 1.0) -> None:
        """
        Increments the value by the given amount.
        By default it will be 1.
        value must be >= 0.
        """
        self._raise_if_cannot_observe()
        if value < 0:
            raise ValueError(f"Counter increase value ({value}) must be >= 0")

        assert self._metric_value_backend is not None
        self._metric_value_backend.inc(value)

    @contextmanager
    def count_exceptions(
        self, exceptions: type[Exception] | tuple[Exception] | None = None
    ) -> Generator[None, None, None]:
        """
        Will count and reraise raised exceptions.
        It is possibly to specify which exceptions to track.
        """
        self._raise_if_cannot_observe()
        if exceptions is None:
            exceptions = Exception

        try:
            yield
        except exceptions:  # type: ignore
            self.inc()
            raise

    def __call__(
        self,
        func: Callable | None = None,
        exceptions: type[Exception] | tuple[Exception] | None = None,
    ) -> Callable:
        """
        When called acts as a decorator counting exceptions raised.
        """
        if func is None:
            return functools.partial(self.__call__, exceptions=exceptions)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):  # type: ignore
            with self.count_exceptions(exceptions):
                return func(*args, **kwargs)

        return wrapper

    def collect(self) -> Iterable[Sample]:
        self._raise_if_cannot_observe()
        assert self._metric_value_backend is not None
        sample = Sample("", self._labels, self._metric_value_backend.get())
        return (self._add_default_labels_to_sample(sample),)


class Gauge(_Metric):
    type_: MetricType = MetricType.GAUGE

    def inc(self, value: float = 1.0) -> None:
        """
        Increments the value by the given amount.
        By default it will be 1.
        """
        self._raise_if_cannot_observe()
        assert self._metric_value_backend is not None
        self._metric_value_backend.inc(value)

    def dec(self, value: float = 1.0) -> None:
        """
        Decrements the value by the given amount.
        By default it will be 1.
        """
        self._raise_if_cannot_observe()
        assert self._metric_value_backend is not None
        self._metric_value_backend.dec(value)

    def set(self, value: float) -> None:
        """
        Set the value to the given amount.
        """
        self._raise_if_cannot_observe()
        assert self._metric_value_backend is not None
        self._metric_value_backend.set(value)

    def set_to_current_time(self) -> None:
        """Set the value to the current unix timestamp."""
        self._raise_if_cannot_observe()
        assert self._metric_value_backend is not None
        self._metric_value_backend.set(time.time())

    @contextmanager
    def track_inprogress(self) -> Generator[None, None, None]:
        """
        Will increase the gauge value when entered and decrease it when exited.
        """
        self._raise_if_cannot_observe()
        self.inc()
        yield
        self.dec()

    def __call__(self, func: Callable | None = None, track_inprogress: bool = False) -> Callable:
        """
        When called acts as a decorator tracking the time taken by
        the wrapped function.

        If passing the parameter `track_inprogress` as `True` it will instead increase while the
        function is running and will decrease when it's finished.
        """
        if func is None:
            return functools.partial(self.__call__, track_inprogress=track_inprogress)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):  # type: ignore
            if track_inprogress:
                with self.track_inprogress():
                    return func(*args, **kwargs)
            else:
                with self.time():
                    return func(*args, **kwargs)

        return wrapper

    @contextmanager
    def time(self) -> Generator[None, None, None]:
        """
        Times the duration inside of it and sets the value.
        """
        self._raise_if_cannot_observe()
        start = time.perf_counter()
        yield
        self.set(time.perf_counter() - start)

    def collect(self) -> Iterable[Sample]:
        self._raise_if_cannot_observe()
        assert self._metric_value_backend is not None
        sample = Sample("", self._labels, self._metric_value_backend.get())
        return (self._add_default_labels_to_sample(sample),)


class Histogram(_Metric):
    type_: MetricType = MetricType.HISTOGRAM
    # Default buckets are tailored to broadly measure the response time (in seconds) of a network
    # service. Most likely you will be required to define buckets customized to your use case.
    # Values taken from the golang/rust client.
    DEFAULT_BUCKETS = (0.005, 0.01, 0.025, 0.05, 0.1, 0.25, 0.5, 1, 2.5, 5, 10)

    def __init__(
        self,
        name: str,
        description: str,
        required_labels: Sequence[str] | None = None,
        labels: Labels | None = None,
        default_labels: Labels | None = None,
        registry: Registry | None = REGISTRY,
        collector: _MetricCollector | None = None,
        buckets: Sequence[float] = DEFAULT_BUCKETS,
    ) -> None:
        super().__init__(
            name,
            description,
            required_labels,
            labels,
            default_labels,
            registry,
            collector,
        )

        # buckets

        # if buckets is None or an empty sequence default buckets will be used
        if not buckets:
            buckets = list(self.DEFAULT_BUCKETS)
        else:
            buckets = list(buckets)

        sorted_buckets = sorted(buckets)
        if buckets != sorted_buckets:
            raise BucketException(
                f"buckets values are not in sorted order. {buckets} != {sorted_buckets}"
            )

        # +Inf is required so we always add it
        last_bucket_value = buckets[-1]
        if isinstance(last_bucket_value, InfFloat):
            pass
        elif last_bucket_value == float("inf"):
            buckets[-1] = InfFloat("inf")
        else:
            buckets.append(InfFloat("inf"))

        self._upper_bounds = buckets

        # create bucket values
        self._buckets = None
        self._sum = None
        self._count = None
        if self._can_observe:
            self._buckets = []

            # this will be added just to the default name on the redis backend but it is
            # fine for now as it's the only one. Might require a more robust way in the future.
            self._sum = get_backend(self, histogram_bucket="sum")
            self._count = get_backend(self, histogram_bucket="count")

            for bucket in self._upper_bounds:
                self._buckets.append(get_backend(self, histogram_bucket=str(bucket)))

    def observe(self, value: float) -> None:
        """
        Observe the given value.
        Value can be negative, in that case the rate function will be less useful so
        it's better to consider using two histograms for positive and negative values.
        """
        self._raise_if_cannot_observe()
        assert self._sum is not None
        self._sum.inc(value)

        for i, bound in enumerate(self._upper_bounds):
            if value <= bound:
                self._buckets[i].inc(1)

        assert self._count is not None
        self._count.inc(1)

    @contextmanager
    def time(self) -> Generator[None, None, None]:
        """
        Times the duration inside of it and sets the value.
        """
        self._raise_if_cannot_observe()
        start = time.perf_counter()
        yield
        self.observe(time.perf_counter() - start)

    def __call__(self, func: Callable) -> Callable:
        """
        When called acts as a decorator tracking the time taken by
        the wrapped function.
        """

        @functools.wraps(func)
        def wrapper(*args, **kwargs):  # type: ignore
            with self.time():
                return func(*args, **kwargs)

        return wrapper

    def collect(self) -> Iterable[Sample]:
        self._raise_if_cannot_observe()
        samples = []
        for i, bound in enumerate(self._upper_bounds):
            bucket_labels = self._labels.copy() if self._labels else {}
            bucket_labels["le"] = str(bound)
            sample = Sample("_bucket", bucket_labels, self._buckets[i].get())
            samples.append(sample)

        assert self._sum is not None
        assert self._count is not None
        samples.append(Sample("_sum", self._labels, self._sum.get()))
        samples.append(Sample("_count", self._labels, self._count.get()))

        return (self._add_default_labels_to_sample(sample) for sample in samples)


# maybe just go with the typing alias
@dataclass
class Label:
    name: str
    value: str



from logging import getLogger
from threading import Lock
from typing import Iterable, Protocol

from pytheus.utils import MetricType

logger = getLogger(__name__)


class Collector(Protocol):
    name: str
    description: str
    type_: MetricType

    def collect(self) -> Iterable:
        ...


class Registry(Protocol):
    prefix: str | None

    def register(self, collector: Collector) -> None:
        ...

    def unregister(self, collector: Collector) -> None:
        ...

    def collect(self) -> Iterable:
        ...


class CollectorRegistry:
    def __init__(self, prefix: str | None = None) -> None:
        self._lock = Lock()
        self.prefix = prefix
        self._collectors: dict[str, Collector] = {}

    def register(self, collector: Collector) -> None:
        with self._lock:
            if collector.name in self._collectors:
                logger.warning(
                    f"collector with name '{collector.name}' already registered."
                    " Keeping previous entry"
                )
                return
            # if it has not this field it's a custom collector, I do not like this approach but
            # let's get it working
            if not hasattr(collector, "type_"):
                for _collector in collector.collect():
                    if _collector.name in self._collectors:
                        logger.warning(
                            f"CustomCollector: {collector.name} - collector with name"
                            f" '{_collector.name}' already registered. Ignoring custom collector"
                        )
                        return

            # NOTE: in case the user is adding directly a metric to the registry
            # we need to add the _MetricCollector to the registry. There might be a better way
            # to d this but it should work for now.
            if hasattr(collector, "_collector"):
                collector = collector._collector
            self._collectors[collector.name] = collector

    def unregister(self, collector: Collector) -> None:
        with self._lock:
            if collector.name not in self._collectors:
                logger.warning(f"no collector found with name '{collector.name}'")
                return
            del self._collectors[collector.name]

    def collect(self) -> Iterable[Collector]:
        yield from self._collectors.values()


class CollectorRegistryProxy:
    def __init__(self, registry: Registry | None = None) -> None:
        self._registry = registry or CollectorRegistry()
        self.prefix = self._registry.prefix

    def set_registry(self, registry: Registry) -> None:
        self._registry = registry
        self.prefix = self._registry.prefix

    def register(self, collector: Collector) -> None:
        self._registry.register(collector)

    def unregister(self, collector: Collector) -> None:
        self._registry.unregister(collector)

    def collect(self) -> Iterable:
        return self._registry.collect()


REGISTRY = CollectorRegistryProxy()
